//-----------------------------------------------------------------------------
// Datei: RevoSoundEngine.cpp
// Beschreibung: Enthlt Funktionen zum Abspielen von Sounds.
//-----------------------------------------------------------------------------


#include <windows.h>
#include <mmsystem.h>
#include "revo3d.h"


static int MP3Index;




RevoSoundEngine::RevoSoundEngine ()
{
	pDSound = NULL;
	pBufPrimary = NULL;
	p3DListener = NULL;
}

RevoSoundEngine::~RevoSoundEngine ()
{
	Cleanup ();
}

//-----------------------------------------------------------------------------
// Funktion: Initialize
// Beschreibung: Initsialisiert die RevoSoundEngine.
// Parameter: hWnd (die hWnd der Hauptengine, mu Revo3dEngine.GethWnd sein
// falls sie verwendet wird).
// Bemerkung: Die RevoSoundEngine verwendet Meter als Maeinheit.
// Setzt als Default-Eigenschaften die 3d-Zuhrer-Position auf (0, 0, 0) und die
// 3d-Zuhrer-Geschwindigkeit auf (0, 0, 0).
//-----------------------------------------------------------------------------
bool RevoSoundEngine::Initialize (HWND hWnd)
{
	char WinFile[MAX_PATH + 256];
	int l;
	DSBUFFERDESC BufDesc;

	Cleanup ();

	//DirectSound erstellen
	if (FAILED (DirectSoundCreate8 (NULL, &pDSound, NULL)))
		return (false);
	pDSound->SetCooperativeLevel (hWnd, DSSCL_EXCLUSIVE);

	//3D-Listener erstellen
	ZeroMemory (&BufDesc, sizeof (DSBUFFERDESC));
	BufDesc.dwSize = sizeof (DSBUFFERDESC);
	BufDesc.dwFlags = DSBCAPS_CTRL3D | DSBCAPS_PRIMARYBUFFER;
	if (FAILED (pDSound->CreateSoundBuffer (&BufDesc, &pBufPrimary, NULL)))
		return (false);
	if (FAILED (pBufPrimary->QueryInterface (IID_IDirectSound3DListener, (void **) &p3DListener)))
		return (false);
	p3DListener->SetPosition (0, 0, 0, DS3D_DEFERRED);
	p3DListener->SetVelocity (0, 0, 0, DS3D_DEFERRED);
	p3DListener->CommitDeferredSettings ();

	//MP3-"Player" erstellen
	l = GetWindowsDirectory (WinFile, MAX_PATH);
	if (WinFile[l - 1] != '\\')
	{
		WinFile[l] = '\\';
		WinFile[l + 1] = '\0';
	}
	strcat (WinFile, "system.ini");
	WritePrivateProfileString ("MCI", "MPEGVideo", "mciqtz.drv", WinFile);

	return (true);
}

//-----------------------------------------------------------------------------
// Funktion: Cleanup
// Beschreibung: Terminiert die RevoSoundEngine.
//-----------------------------------------------------------------------------
void RevoSoundEngine::Cleanup ()
{
	if (p3DListener)
	{
		p3DListener->Release ();
		p3DListener = NULL;
	}
	if (pBufPrimary)
	{
		pBufPrimary->Release ();
		pBufPrimary = NULL;
	}
	if (pDSound)
	{
		pDSound->Release ();
		pDSound = NULL;
	}
}



//-----------------------------------------------------------------------------
// Funktion: SetListener3dPosition
// Beschreibung: Setzt die Position des Zuhrers in der 3d-Welt (meistens die
// Position der Kamera).
// Parameter: Pos (die Position, in Meter), LookAt (der Punkt wohin der Zuhrer sieht, in Meter)
// Bemerkung: Diese Funktion muss mit Update3dSounds bernommen werden.
//-----------------------------------------------------------------------------
void RevoSoundEngine::SetListener3dPosition (D3DXVECTOR3 *Pos, D3DXVECTOR3 *LookAt)
{
	if (!p3DListener || !Pos || !LookAt)
		return;
	if (Pos->x == LookAt->x && Pos->y == LookAt->y && Pos->z == LookAt->z)
		return;

	D3DXVECTOR3 Delta, DeltaCube, FrontVec, TopVec;
	float DeltaCubeSum, HypSum, sinx, cosx, siny, cosy;

	//Position selbst setzen
	Position = *Pos;
	p3DListener->SetPosition (Pos->x, Pos->y, Pos->z, DS3D_DEFERRED);
	//Top-Vektor erstellen
	D3DXVec3Subtract (&Delta, LookAt, Pos);
	DeltaCube.x = Delta.x * Delta.x;
	DeltaCube.y = Delta.y * Delta.y;
	DeltaCube.z = Delta.z * Delta.z;
	DeltaCubeSum = DeltaCube.x + DeltaCube.y + DeltaCube.z;
	//sin und cos der beiden Winkeln berechnen
	sinx = Delta.y / (float) sqrt ((double) DeltaCubeSum);
	cosx = (float) sqrt ((double) (1 - DeltaCube.y / DeltaCubeSum));
	if (DeltaCube.x + DeltaCube.z)
	{
		HypSum = (float) sqrt ((double) (DeltaCube.x + DeltaCube.z));
		siny = Delta.x / HypSum;
		cosy = Delta.z / HypSum;
	}
	else
	{
		siny = 0;
		cosy = 1;
	}
	TopVec = Vector3dMake (-sinx * siny, cosx, -sinx * cosy);
	D3DXVec3Normalize (&TopVec, &TopVec);
	//Front-Vektor erstellen
	D3DXVec3Subtract (&FrontVec, LookAt, Pos);
	D3DXVec3Normalize (&FrontVec, &FrontVec);
	p3DListener->SetOrientation (FrontVec.x, FrontVec.y, FrontVec.z, TopVec.x, TopVec.y, TopVec.z, DS3D_DEFERRED);
}

//-----------------------------------------------------------------------------
// Funktion: MoveListener3dPosition
// Beschreibung: Verschiebt die Position des Zuhrers in der 3d-Welt.
// Parameter: MoveDelta (die Verschiebung, relativ zum aktuellen Standpunkt, in Meter),
// LookAt (der Punkt wohin der Zuhrer sieht, in Meter)
// Bemerkung: Diese Funktion muss mit Update3dSounds bernommen werden.
//-----------------------------------------------------------------------------
void RevoSoundEngine::MoveListener3dPosition (D3DXVECTOR3 *MoveDelta, D3DXVECTOR3 *LookAt)
{
	if (!p3DListener || !MoveDelta || !LookAt)
		return;
	D3DXVec3Add (&Position, &Position, MoveDelta);
	SetListener3dPosition (&Position, LookAt);
}

//-----------------------------------------------------------------------------
// Funktion: SetListenerSpeed
// Beschreibung: Wird ausschlielich zur Berechnung des Doppler-Effektes verwendet.
// Diese Funktion hat daher keinen Einfluss auf die wirkliche Zuhrerposition oder
// -geschwindigkeit.
// Parameter: Speed (die Geschwindigkeit als Vektor, in Meter pro Sekunde)
// Bemerkung: Diese Funktion muss mit Update3dSounds bernommen werden.
//-----------------------------------------------------------------------------
void RevoSoundEngine::SetListenerSpeed (D3DXVECTOR3 *Speed)
{
	if (!p3DListener || !Speed)
		return;
	p3DListener->SetVelocity (Speed->x, Speed->y, Speed->z, DS3D_DEFERRED);
}



//-----------------------------------------------------------------------------
// Funktion: Play2dSound
// Beschreibung: Spielt einen 2d-Sound ab.
// Parameter: Sound (der Sound), Looping (ob der Sound immer wiederholt werden soll)
//-----------------------------------------------------------------------------
void RevoSoundEngine::Play2dSound (Revo2dSound *Sound, bool Looping)
{
	if (!Sound)
		return;

	LPDIRECTSOUNDBUFFER Buffer = Sound->pSoundBuf;

	if (!Buffer)
		return;
	if (Looping)
		Buffer->Play (0, 0, DSBPLAY_LOOPING);
	else
		Buffer->Play (0, 0, 0);
}

//-----------------------------------------------------------------------------
// Funktion: Stop2dSound
// Beschreibung: Stoppt das Abspielen eines 2d-Sounds.
// Parameter: Sound (der Sound)
//-----------------------------------------------------------------------------
void RevoSoundEngine::Stop2dSound (Revo2dSound *Sound)
{
	if (!Sound)
		return;

	LPDIRECTSOUNDBUFFER Buffer = Sound->pSoundBuf;

	if (!Buffer)
		return;
	Buffer->Stop ();
}



//-----------------------------------------------------------------------------
// Funktion: Play3dSound
// Beschreibung: Spielt einen 3d-Sound ab.
// Parameter: Sound (der Sound), Looping (ob der Sound immer wiederholt werden soll)
//-----------------------------------------------------------------------------
void RevoSoundEngine::Play3dSound (Revo3dSound *Sound, bool Looping)
{
	if (!Sound)
		return;

	LPDIRECTSOUNDBUFFER Buffer = Sound->pSoundBuf;

	if (!Buffer)
		return;	
	if (Looping)
		Buffer->Play (0, 0, DSBPLAY_LOOPING);
	else
		Buffer->Play (0, 0, 0);
}

//-----------------------------------------------------------------------------
// Funktion: Stop3dSound
// Beschreibung: Stoppt das Abspielen eines 3d-Sounds.
// Parameter: Sound (der Sound)
//-----------------------------------------------------------------------------
void RevoSoundEngine::Stop3dSound (Revo3dSound *Sound)
{
	if (!Sound)
		return;

	LPDIRECTSOUNDBUFFER Buffer = Sound->pSoundBuf;

	if (!Buffer)
		return;	
	Buffer->Stop ();
}



//-----------------------------------------------------------------------------
// Funktion: PlayMP3
// Beschreibung: Spielt ein MP3 ab.
// Parameter: MP3 (das MP3), Looping (ob das MP3 immer wiederholt werden soll)
//-----------------------------------------------------------------------------
void RevoSoundEngine::PlayMP3 (RevoMP3 *MP3, bool Looping)
{
	if (!MP3)
		return;

	char cmd[256];

	strcpy (cmd, "play mpeg");
	strcat (cmd, MP3->sMP3Ind);
	if (Looping)
		strcat (cmd, " repeat");
	mciSendString (cmd, NULL, 0, 0);
}

//-----------------------------------------------------------------------------
// Funktion: StopMP3
// Beschreibung: Stoppt das Abspielen eines MP3s.
// Parameter: MP3 (das MP3)
//-----------------------------------------------------------------------------
void RevoSoundEngine::StopMP3 (RevoMP3 *MP3)
{
	if (!MP3)
		return;

	char cmd[256];

	strcpy (cmd, "stop mpeg");
	strcat (cmd, MP3->sMP3Ind);
	mciSendString (cmd, NULL, 0, 0);
}



//-----------------------------------------------------------------------------
// Funktion: Update3dSounds
// Beschreibung: bernimmt alle vorgenommenen nderungen an den 3d-Sounds und
// des Listeners, die aktualisiert werden mssen. 
//-----------------------------------------------------------------------------
void RevoSoundEngine::Update3dSounds ()
{
	if (!p3DListener)
		return;
	p3DListener->CommitDeferredSettings ();
}




















Revo2dSound::Revo2dSound ()
{
	pSoundBuf = NULL;
}

Revo2dSound &Revo2dSound::operator= (Revo2dSound &pOrig)
{
	if (&pOrig == this)
		return (*this);	

	void *LockedBuf, *LockedBufOrig;
	DWORD LockedBufSize, LockedBufSizeOrig;

	if (BufSize == pOrig.BufSize && pSoundBuf && pOrig.pSoundBuf)
	{
		//Buffer kopieren
		pSoundBuf->Lock (0, BufSize, &LockedBuf, &LockedBufSize, NULL, NULL, 0);
		pOrig.pSoundBuf->Lock (0, BufSize, &LockedBufOrig, &LockedBufSizeOrig, NULL, NULL, 0);
		memcpy (LockedBuf, LockedBufOrig, BufSize);
		pOrig.pSoundBuf->Unlock (LockedBufOrig, LockedBufSizeOrig, NULL, 0);
		pSoundBuf->Unlock (LockedBuf, LockedBufSize, NULL, 0);
		return (*this);
	}
	//Buffer neu erstellen
	pDSound = pOrig.pDSound;
	LoadFromFile (pOrig.File);
	return (*this);
}

Revo2dSound::~Revo2dSound ()
{
	Cleanup ();
}

//-----------------------------------------------------------------------------
// Funktion: Initialize
// Beschreibung: Initsialisiert den Revo2dSound.
// Parameter: Snd (die Soundengine)
//-----------------------------------------------------------------------------
void Revo2dSound::Initialize (RevoSoundEngine *Snd)
{
	Cleanup ();
	pDSound = Snd->pDSound;
}

//-----------------------------------------------------------------------------
// Funktion: Cleanup
// Beschreibung: Terminiert den Revo2dSound.
//-----------------------------------------------------------------------------
void Revo2dSound::Cleanup ()
{
	if (pSoundBuf)
	{
		pSoundBuf->Stop ();
		pSoundBuf->Release ();
		pSoundBuf = NULL;
	}
}



//-----------------------------------------------------------------------------
// Funktion: LoadFromFile
// Beschreibung: Ldt einen Sound von einer wav-Datei.
// Parameter: File (der Pfad der wav-Datei)
// Bemerkung: Setzt als Default-Eigenschaften die Abspielposition auf 0 und die Lautstrke auf 1.
//-----------------------------------------------------------------------------
bool Revo2dSound::LoadFromFile (char *File)
{
	if (!pDSound || !File)
		return (false);

	DSBUFFERDESC BufDesc;
	WAVEFORMATEX *pFormat;
	void *LockedBuf;
	DWORD LockedBufSize, RdSize;
	CWaveFile RdFile;


	Cleanup ();				//frheren Buffer lschen, falls vorhanden

	//Wave-File ffnen
	RdFile.Open (File, NULL);
	if (!RdFile.GetSize ())
	{
		RdFile.Close ();
		return (false);
	}
	pFormat = RdFile.GetFormat ();
	if (!pFormat || pFormat->wFormatTag != WAVE_FORMAT_PCM)
	{
		RdFile.Close ();
		return (false);
	}
	BufSize = RdFile.GetSize ();

	//Buffer erstellen
	ZeroMemory (&BufDesc, sizeof (DSBUFFERDESC));
	BufDesc.dwSize = sizeof (DSBUFFERDESC);
	BufDesc.dwFlags = DSBCAPS_CTRLVOLUME;
	BufDesc.dwBufferBytes = BufSize;
	BufDesc.guid3DAlgorithm = GUID_NULL;
	BufDesc.lpwfxFormat = RdFile.m_pwfx;
	if (FAILED (pDSound->CreateSoundBuffer (&BufDesc, &pSoundBuf, NULL)))
	{
		RdFile.Close ();
		return (false);
	}

	//Buffer mit Sound von der Datei fllen und Datei schlieen
	pSoundBuf->Lock (0, BufSize, &LockedBuf, &LockedBufSize, NULL, NULL, 0);
	RdFile.ResetFile ();
	RdFile.Read ((BYTE *) LockedBuf, LockedBufSize, &RdSize);
	pSoundBuf->Unlock (LockedBuf, LockedBufSize, NULL, 0);
	RdFile.Close ();

	strcpy (this->File, File);

	return (true);
}



//-----------------------------------------------------------------------------
// Funktion: SetPlayPosition
// Beschreibung: Setzt die Position von der weg der Sound abgespielt wird.
// Parameter: Pos (die Position, von 0 (= Anfang) bis 1 (= Ende))
//-----------------------------------------------------------------------------
void Revo2dSound::SetPlayPosition (float Pos)
{
	if (!pSoundBuf)
		return;
	pSoundBuf->SetCurrentPosition ((DWORD) (Pos * BufSize));
}



//-----------------------------------------------------------------------------
// Funktion: SetVolume
// Beschreibung: Setzt die Lautstrke fr den Sound.
// Parameter: Vol (die Lautstrke, von 0 (= Stille) bis 1 (= originale Lautstrke))
//-----------------------------------------------------------------------------
void Revo2dSound::SetVolume (float Vol)
{
	if (!pSoundBuf)
		return;
	pSoundBuf->SetVolume ((DWORD) (DSBVOLUME_MIN + Vol * VolDiff));
}




















RevoMP3::RevoMP3 ()
{
	itoa (MP3Index, sMP3Ind, 10);
	MP3Index++;
}

RevoMP3 &RevoMP3::operator= (RevoMP3 &pOrig)
{
	if (&pOrig != this)
		LoadFromFile (pOrig.File);
	return (*this);
}

RevoMP3::~RevoMP3 ()
{
	Cleanup ();
}

//-----------------------------------------------------------------------------
// Funktion: Cleanup
// Beschreibung: Terminiert das RevoMP3.
//-----------------------------------------------------------------------------
void RevoMP3::Cleanup ()
{
	char cmd[256];

	strcpy (cmd, "close mpeg");
	strcat (cmd, sMP3Ind);
	mciSendString (cmd, NULL, 0, 0);
}



//-----------------------------------------------------------------------------
// Funktion: LoadFromFile
// Beschreibung: Ldt eine mp3-Datei.
// Parameter: File (der Pfad der mp3-Datei)
// Bemerkung: Setzt als Default-Eigenschaften die Abspielposition auf 0.
//-----------------------------------------------------------------------------
bool RevoMP3::LoadFromFile (char *File)
{
	char ShortFile[MAX_PATH + 1], cmd[MAX_PATH + 256], slen[256];

	//Frhere Datei schlieen, falls geffnet
	Cleanup ();
	//MP3-Datei ffnen
	GetShortPathName (File, ShortFile, MAX_PATH);
	strcpy (cmd, "open ");
	strcat (cmd, ShortFile);
	strcat (cmd, " type MPEGVideo alias mpeg");
	strcat (cmd, sMP3Ind);
	if (mciSendString (cmd, NULL, 0, 0))
		return (false);
	//Zeitformat auf Frames setzen
	strcpy (cmd, "set mpeg");
	strcat (cmd, sMP3Ind);
	strcat (cmd, " time format frames");
	mciSendString (cmd, NULL, 0, 0);
	//Lnge ermitteln
	strcpy (cmd, "status mpeg");
	strcat (cmd, sMP3Ind);
	strcat (cmd, " length");
	mciSendString (cmd, slen, 255, 0);
	MP3Len = atol (slen);
	//Abspielposition auf den Anfang setzen
	SetPlayPosition (0);

	strcpy (this->File, File);

	return (true);
}



//-----------------------------------------------------------------------------
// Funktion: SetPlayPosition
// Beschreibung: Setzt die Position von der weg das MP3 abgespielt wird.
// Parameter: Pos (die Position, von 0 (= Anfang) bis 1 (= Ende))
//-----------------------------------------------------------------------------
void RevoMP3::SetPlayPosition (float Pos)
{
	char cmd[256], sSeekPos[256];
	long SeekPos;

	SeekPos = (long) (MP3Len * Pos);
	strcpy (cmd, "seek mpeg");
	strcat (cmd, sMP3Ind);
	strcat (cmd, " to ");
	strcat (cmd, ltoa (SeekPos, sSeekPos, 10));
	mciSendString (cmd, NULL, 0, 0);
}




















Revo3dSound::Revo3dSound ()
{
	pSoundBuf = NULL;
	pSound3DBuf = NULL;
}

Revo3dSound &Revo3dSound::operator= (Revo3dSound &pOrig)
{
	if (&pOrig == this)
		return (*this);	

	void *LockedBuf, *LockedBufOrig;
	DWORD LockedBufSize, LockedBufSizeOrig;

	Position = pOrig.Position;
	if (BufSize == pOrig.BufSize && pSoundBuf && pOrig.pSoundBuf)
	{
		//Buffer kopieren
		pSoundBuf->Lock (0, BufSize, &LockedBuf, &LockedBufSize, NULL, NULL, 0);
		pOrig.pSoundBuf->Lock (0, BufSize, &LockedBufOrig, &LockedBufSizeOrig, NULL, NULL, 0);
		memcpy (LockedBuf, LockedBufOrig, BufSize);
		pOrig.pSoundBuf->Unlock (LockedBufOrig, LockedBufSizeOrig, NULL, 0);
		pSoundBuf->Unlock (LockedBuf, LockedBufSize, NULL, 0);
		return (*this);
	}
	//Buffer neu erstellen
	pDSound = pOrig.pDSound;
	LoadFromFile (pOrig.File);
	return (*this);
}

Revo3dSound::~Revo3dSound ()
{
	Cleanup ();
}

//-----------------------------------------------------------------------------
// Funktion: Initialize
// Beschreibung: Initsialisiert den Revo3dSound.
// Parameter: Snd (die Soundengine)
//-----------------------------------------------------------------------------
void Revo3dSound::Initialize (RevoSoundEngine *Snd)
{
	Cleanup ();
	pDSound = Snd->pDSound;
}

//-----------------------------------------------------------------------------
// Funktion: Cleanup
// Beschreibung: Terminiert den Revo3dSound.
//-----------------------------------------------------------------------------
void Revo3dSound::Cleanup ()
{
	if (pSoundBuf)
		pSoundBuf->Stop ();
	if (pSound3DBuf)
	{
		pSound3DBuf->Release ();
		pSound3DBuf = NULL;
	}
	if (pSoundBuf)
	{
		pSoundBuf->Release ();
		pSoundBuf = NULL;
	}
}



//-----------------------------------------------------------------------------
// Funktion: LoadFromFile
// Beschreibung: Ldt einen Sound von einer wav-Datei.
// Parameter: File (der Pfad der wav-Datei)
// Bemerkung: Fr 3d-Sounds muss die wav-Datei mono sein.
// Setzt als Default-Eigenschaften die Abspielposition auf 0, die Lautstrke auf 1,
// die 3d-Position auf (0, 0, 0) und die 3d-Geschwindigkeit auf (0, 0, 0).
//-----------------------------------------------------------------------------
bool Revo3dSound::LoadFromFile (char *File)
{
	if (!pDSound || !File)
		return (false);

	DSBUFFERDESC BufDesc;
	WAVEFORMATEX *pFormat;
	void *LockedBuf;
	DWORD LockedBufSize, RdSize;
	CWaveFile RdFile;


	Cleanup ();				//frheren Buffer lschen, falls vorhanden

	//Wave-File ffnen
	RdFile.Open (File, NULL);
	if (!RdFile.GetSize ())
	{
		RdFile.Close ();
		return (false);
	}
	pFormat = RdFile.GetFormat ();
	if (!pFormat || pFormat->wFormatTag != WAVE_FORMAT_PCM || pFormat->nChannels > 1)
	{
		RdFile.Close ();
		return (false);
	}
	BufSize = RdFile.GetSize ();

	//Buffer erstellen
	ZeroMemory (&BufDesc, sizeof (DSBUFFERDESC));
	BufDesc.dwSize = sizeof (DSBUFFERDESC);
	BufDesc.dwFlags = DSBCAPS_CTRL3D | DSBCAPS_CTRLVOLUME;
	BufDesc.dwBufferBytes = BufSize;
	BufDesc.guid3DAlgorithm = DS3DALG_HRTF_FULL;
	BufDesc.lpwfxFormat = RdFile.m_pwfx;
	if (FAILED (pDSound->CreateSoundBuffer (&BufDesc, &pSoundBuf, NULL)))
	{
		RdFile.Close ();
		return (false);
	}

	//Buffer mit Sound von der Datei fllen und Datei schlieen
	pSoundBuf->Lock (0, BufSize, &LockedBuf, &LockedBufSize, NULL, NULL, 0);
	RdFile.ResetFile ();
	RdFile.Read ((BYTE *) LockedBuf, LockedBufSize, &RdSize);
	pSoundBuf->Unlock (LockedBuf, LockedBufSize, NULL, 0);
	RdFile.Close ();

	//3D-Buffer erstellen und initsialisieren
	if (FAILED (pSoundBuf->QueryInterface (IID_IDirectSound3DBuffer, (void **) &pSound3DBuf)))
		return (false);
	Set3dPosition (&Vector3dMake (0, 0, 0));
	SetSpeed (&Vector3dMake (0, 0, 0));

	strcpy (this->File, File);

	return (true);
}



//-----------------------------------------------------------------------------
// Funktion: SetDistances
// Beschreibung: Setzt die minimale und maximale Distanz fr den Sound. Die minimale
// Distanz beschreibt, ab welcher Entfernung der Sound nicht mehr lauter wird. Die maximale
// Distanz beschreibt, ab welcher Entfernung der Sound nicht mehr leiser wird. Dies ist fr
// die Optimierung bei der Berechnung wichtig und sollte immer gesetzt werden.
// Parameter: MinDist (die minimale Distanz), MaxDist (die maximale Distanz)
// Bemerkung: Diese Funktion muss mit Update3dSounds bernommen werden.
//-----------------------------------------------------------------------------
void Revo3dSound::SetDistances (float MinDist, float MaxDist)
{
	if (!pSound3DBuf)
		return;
	pSound3DBuf->SetMinDistance (MinDist, DS3D_DEFERRED);
	pSound3DBuf->SetMaxDistance (MaxDist, DS3D_DEFERRED);
}



//-----------------------------------------------------------------------------
// Funktion: SetPlayPosition
// Beschreibung: Setzt die Position von der weg der Sound abgespielt wird.
// Parameter: Pos (die Position, von 0 (= Anfang) bis 1 (= Ende))
//-----------------------------------------------------------------------------
void Revo3dSound::SetPlayPosition (float Pos)
{
	if (!pSoundBuf)
		return;
	pSoundBuf->SetCurrentPosition ((DWORD) (Pos * BufSize));
}



//-----------------------------------------------------------------------------
// Funktion: Set3dPosition
// Beschreibung: Setzt die Position des Sounds in der 3d-Welt.
// Parameter: Pos (die Position, in Meter)
// Bemerkung: Diese Funktion muss mit Update3dSounds bernommen werden.
//-----------------------------------------------------------------------------
void Revo3dSound::Set3dPosition (D3DXVECTOR3 *Pos)
{
	if (!pSound3DBuf || !Pos)
		return;
	Position = *Pos;
	pSound3DBuf->SetPosition (Pos->x, Pos->y, Pos->z, DS3D_DEFERRED);
}

//-----------------------------------------------------------------------------
// Funktion: Move3dPosition
// Beschreibung: Verschiebt die Position des Sounds in der 3d-Welt.
// Parameter: MoveDelta (die Verschiebung, relativ zum aktuellen Standpunkt, in Meter)
// Bemerkung: Diese Funktion muss mit Update3dSounds bernommen werden.
//-----------------------------------------------------------------------------
void Revo3dSound::Move3dPosition (D3DXVECTOR3 *MoveDelta)
{
	if (!pSound3DBuf || !MoveDelta)
		return;
	D3DXVec3Add (&Position, &Position, MoveDelta);
	pSound3DBuf->SetPosition (Position.x, Position.y, Position.z, DS3D_DEFERRED);
}

//-----------------------------------------------------------------------------
// Funktion: SetSpeed
// Beschreibung: Wird ausschlielich zur Berechnung des Doppler-Effektes verwendet.
// Diese Funktion hat daher keinen Einfluss auf die wirkliche Soundposition oder
// -geschwindigkeit.
// Parameter: Speed (die Geschwindigkeit als Vektor, in Meter pro Sekunde)
// Bemerkung: Diese Funktion muss mit Update3dSounds bernommen werden.
//-----------------------------------------------------------------------------
void Revo3dSound::SetSpeed (D3DXVECTOR3 *Speed)
{
	if (!pSound3DBuf || !Speed)
		return;
	pSound3DBuf->SetVelocity (Speed->x, Speed->y, Speed->z, DS3D_DEFERRED);
}



//-----------------------------------------------------------------------------
// Funktion: SetVolume
// Beschreibung: Setzt die Lautstrke fr den Sound.
// Parameter: Vol (die Lautstrke, von 0 (= Stille) bis 1 (= originale Lautstrke))
// Bemerkung: 3d-Sounds werden automatisch leiser, wenn sie weiter vom "Zuhrer" entfernt
// sind und lauter wenn sie nher sind. Dieser Volume-Wert bezieht sich nur auf die
// Lautstrke, die von der Schallquelle selbst ausgestrahlt wird.
//-----------------------------------------------------------------------------
void Revo3dSound::SetVolume (float Vol)
{
	int VolDiff = DSBVOLUME_MAX - DSBVOLUME_MIN;

	if (!pSoundBuf)
		return;
	pSoundBuf->SetVolume ((DWORD) (DSBVOLUME_MIN + Vol * VolDiff));
}





















//-----------------------------------------------------------------------------
// Name: CWaveFile::CWaveFile()
// Desc: Constructs the class.  Call Open() to open a wave file for reading.  
//       Then call Read() as needed.  Calling the destructor or Close() 
//       will close the file.  
//-----------------------------------------------------------------------------
CWaveFile::CWaveFile()
{
    m_pwfx    = NULL;
    m_hmmio   = NULL;
    m_dwSize  = 0;
}




//-----------------------------------------------------------------------------
// Name: CWaveFile::~CWaveFile()
// Desc: Destructs the class
//-----------------------------------------------------------------------------
CWaveFile::~CWaveFile()
{
	SAFE_DELETE_ARRAY( m_pwfx );
}




//-----------------------------------------------------------------------------
// Name: CWaveFile::Open()
// Desc: Opens a wave file for reading
//-----------------------------------------------------------------------------
HRESULT CWaveFile::Open( LPTSTR strFileName, WAVEFORMATEX* pwfx )
{
    HRESULT hr;

    if( strFileName == NULL )
        return E_INVALIDARG;
    SAFE_DELETE_ARRAY( m_pwfx );

    m_hmmio = mmioOpen( strFileName, NULL, MMIO_ALLOCBUF | MMIO_READ );

    if( NULL == m_hmmio )
    {
        HRSRC   hResInfo;
        HGLOBAL hResData;
        DWORD   dwSize;
        VOID*   pvRes;

        // Loading it as a file failed, so try it as a resource
        if( NULL == ( hResInfo = FindResource( NULL, strFileName, TEXT("WAVE") ) ) )
        {
            if( NULL == ( hResInfo = FindResource( NULL, strFileName, TEXT("WAV") ) ) )
                return DXTRACE_ERR_NOMSGBOX( TEXT("FindResource"), E_FAIL );
        }

        if( NULL == ( hResData = LoadResource( NULL, hResInfo ) ) )
            return DXTRACE_ERR( TEXT("LoadResource"), E_FAIL );

        if( 0 == ( dwSize = SizeofResource( NULL, hResInfo ) ) ) 
            return DXTRACE_ERR( TEXT("SizeofResource"), E_FAIL );

        if( NULL == ( pvRes = LockResource( hResData ) ) )
            return DXTRACE_ERR( TEXT("LockResource"), E_FAIL );

        CHAR* pData = new CHAR[ dwSize ];
        memcpy( pData, pvRes, dwSize );

        MMIOINFO mmioInfo;
        ZeroMemory( &mmioInfo, sizeof(mmioInfo) );
        mmioInfo.fccIOProc = FOURCC_MEM;
        mmioInfo.cchBuffer = dwSize;
        mmioInfo.pchBuffer = (CHAR*) pData;

        m_hmmio = mmioOpen( NULL, &mmioInfo, MMIO_ALLOCBUF | MMIO_READ );
    }

    if( FAILED( hr = ReadMMIO() ) )
    {
        // ReadMMIO will fail if its an not a wave file
        mmioClose( m_hmmio, 0 );
        return DXTRACE_ERR_NOMSGBOX( TEXT("ReadMMIO"), hr );
    }

    if( FAILED( hr = ResetFile() ) )
        return DXTRACE_ERR( TEXT("ResetFile"), hr );

    // After the reset, the size of the wav file is m_ck.cksize so store it now
    m_dwSize = m_ck.cksize;

    return hr;
}




//-----------------------------------------------------------------------------
// Name: CWaveFile::ReadMMIO()
// Desc: Support function for reading from a multimedia I/O stream.
//       m_hmmio must be valid before calling.  This function uses it to
//       update m_ckRiff, and m_pwfx. 
//-----------------------------------------------------------------------------
HRESULT CWaveFile::ReadMMIO()
{
    MMCKINFO        ckIn;           // chunk info. for general use.
    PCMWAVEFORMAT   pcmWaveFormat;  // Temp PCM structure to load in.       

    m_pwfx = NULL;

    if( ( 0 != mmioDescend( m_hmmio, &m_ckRiff, NULL, 0 ) ) )
        return DXTRACE_ERR( TEXT("mmioDescend"), E_FAIL );

    // Check to make sure this is a valid wave file
    if( (m_ckRiff.ckid != FOURCC_RIFF) ||
        (m_ckRiff.fccType != mmioFOURCC('W', 'A', 'V', 'E') ) )
        return DXTRACE_ERR_NOMSGBOX( TEXT("mmioFOURCC"), E_FAIL ); 

    // Search the input file for for the 'fmt ' chunk.
    ckIn.ckid = mmioFOURCC('f', 'm', 't', ' ');
    if( 0 != mmioDescend( m_hmmio, &ckIn, &m_ckRiff, MMIO_FINDCHUNK ) )
        return DXTRACE_ERR( TEXT("mmioDescend"), E_FAIL );

    // Expect the 'fmt' chunk to be at least as large as <PCMWAVEFORMAT>;
    // if there are extra parameters at the end, we'll ignore them
       if( ckIn.cksize < (LONG) sizeof(PCMWAVEFORMAT) )
           return DXTRACE_ERR( TEXT("sizeof(PCMWAVEFORMAT)"), E_FAIL );

    // Read the 'fmt ' chunk into <pcmWaveFormat>.
    if( mmioRead( m_hmmio, (HPSTR) &pcmWaveFormat, 
                  sizeof(pcmWaveFormat)) != sizeof(pcmWaveFormat) )
        return DXTRACE_ERR( TEXT("mmioRead"), E_FAIL );

    // Allocate the waveformatex, but if its not pcm format, read the next
    // word, and thats how many extra bytes to allocate.
    if( pcmWaveFormat.wf.wFormatTag == WAVE_FORMAT_PCM )
    {
        m_pwfx = (WAVEFORMATEX*)new CHAR[ sizeof(WAVEFORMATEX) ];
        if( NULL == m_pwfx )
            return DXTRACE_ERR( TEXT("m_pwfx"), E_FAIL );

        // Copy the bytes from the pcm structure to the waveformatex structure
        memcpy( m_pwfx, &pcmWaveFormat, sizeof(pcmWaveFormat) );
        m_pwfx->cbSize = 0;
    }
    else
    {
        // Read in length of extra bytes.
        WORD cbExtraBytes = 0L;
        if( mmioRead( m_hmmio, (CHAR*)&cbExtraBytes, sizeof(WORD)) != sizeof(WORD) )
            return DXTRACE_ERR( TEXT("mmioRead"), E_FAIL );

        m_pwfx = (WAVEFORMATEX*)new CHAR[ sizeof(WAVEFORMATEX) + cbExtraBytes ];
        if( NULL == m_pwfx )
            return DXTRACE_ERR( TEXT("new"), E_FAIL );

        // Copy the bytes from the pcm structure to the waveformatex structure
        memcpy( m_pwfx, &pcmWaveFormat, sizeof(pcmWaveFormat) );
        m_pwfx->cbSize = cbExtraBytes;

        // Now, read those extra bytes into the structure, if cbExtraAlloc != 0.
        if( mmioRead( m_hmmio, (CHAR*)(((BYTE*)&(m_pwfx->cbSize))+sizeof(WORD)),
                      cbExtraBytes ) != cbExtraBytes )
        {
            SAFE_DELETE( m_pwfx );
            return DXTRACE_ERR( TEXT("mmioRead"), E_FAIL );
        }
    }

    // Ascend the input file out of the 'fmt ' chunk.
    if( 0 != mmioAscend( m_hmmio, &ckIn, 0 ) )
    {
        SAFE_DELETE( m_pwfx );
        return DXTRACE_ERR( TEXT("mmioAscend"), E_FAIL );
    }

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: CWaveFile::GetSize()
// Desc: Retuns the size of the read access wave file 
//-----------------------------------------------------------------------------
DWORD CWaveFile::GetSize()
{
    return m_dwSize;
}




//-----------------------------------------------------------------------------
// Name: CWaveFile::ResetFile()
// Desc: Resets the internal m_ck pointer so reading starts from the 
//       beginning of the file again 
//-----------------------------------------------------------------------------
HRESULT CWaveFile::ResetFile()
{
    if( m_hmmio == NULL )
        return CO_E_NOTINITIALIZED;

    // Seek to the data
    if( -1 == mmioSeek( m_hmmio, m_ckRiff.dwDataOffset + sizeof(FOURCC),
                    SEEK_SET ) )
        return DXTRACE_ERR( TEXT("mmioSeek"), E_FAIL );

    // Search the input file for the 'data' chunk.
    m_ck.ckid = mmioFOURCC('d', 'a', 't', 'a');
    if( 0 != mmioDescend( m_hmmio, &m_ck, &m_ckRiff, MMIO_FINDCHUNK ) )
      return DXTRACE_ERR( TEXT("mmioDescend"), E_FAIL );
    
    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: CWaveFile::Read()
// Desc: Reads section of data from a wave file into pBuffer and returns 
//       how much read in pdwSizeRead, reading not more than dwSizeToRead.
//       This uses m_ck to determine where to start reading from.  So 
//       subsequent calls will be continue where the last left off unless 
//       Reset() is called.
//-----------------------------------------------------------------------------
HRESULT CWaveFile::Read( BYTE* pBuffer, DWORD dwSizeToRead, DWORD* pdwSizeRead )
{
    MMIOINFO mmioinfoIn; // current status of m_hmmio

    if( m_hmmio == NULL )
        return CO_E_NOTINITIALIZED;
    if( pBuffer == NULL || pdwSizeRead == NULL )
        return E_INVALIDARG;

    if( pdwSizeRead != NULL )
        *pdwSizeRead = 0;

    if( 0 != mmioGetInfo( m_hmmio, &mmioinfoIn, 0 ) )
        return DXTRACE_ERR( TEXT("mmioGetInfo"), E_FAIL );
            
    UINT cbDataIn = dwSizeToRead;
    if( cbDataIn > m_ck.cksize ) 
        cbDataIn = m_ck.cksize;       

    m_ck.cksize -= cbDataIn;

    for( DWORD cT = 0; cT < cbDataIn; cT++ )
    {
        // Copy the bytes from the io to the buffer.
        if( mmioinfoIn.pchNext == mmioinfoIn.pchEndRead )
        {
            if( 0 != mmioAdvance( m_hmmio, &mmioinfoIn, MMIO_READ ) )
                return DXTRACE_ERR( TEXT("mmioAdvance"), E_FAIL );

            if( mmioinfoIn.pchNext == mmioinfoIn.pchEndRead )
                return DXTRACE_ERR( TEXT("mmioinfoIn.pchNext"), E_FAIL );
        }

        // Actual copy.
        *((BYTE*)pBuffer+cT) = *((BYTE*)mmioinfoIn.pchNext);
        mmioinfoIn.pchNext++;
    }

    if( 0 != mmioSetInfo( m_hmmio, &mmioinfoIn, 0 ) )
        return DXTRACE_ERR( TEXT("mmioSetInfo"), E_FAIL );

    if( pdwSizeRead != NULL )
        *pdwSizeRead = cbDataIn;

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: CWaveFile::Close()
// Desc: Closes the wave file 
//-----------------------------------------------------------------------------
HRESULT CWaveFile::Close()
{
    mmioClose( m_hmmio, 0 );
    m_hmmio = NULL;

    return S_OK;
}